import requests
import os
from tqdm import tqdm
import zipfile
import io
from typing import Optional, Union
import shutil

class LoadTools:
    """
    Class dedicated to downloading / loading models and datasets from either the BallDataLab API or specified text files.

    Attributes:
        session (requests.Session): Session object for making requests.
        chunk_size (int): Size of chunks to use when downloading files.
        BDL_MODEL_API (str): Base URL for the BallDataLab model API.
        BDL_DATASET_API (str): Base URL for the BallDataLab dataset API.

    Methods:
        load_model(model_alias: str, model_type: str = 'YOLO', use_bdl_api: Optional[bool] = True) -> str:
            Loads a given baseball computer vision model into the repository.
        load_dataset(dataset_alias: str, use_bdl_api: Optional[bool] = True) -> str:
            Loads a zipped dataset and extracts it to a folder.
        _download_files(url: str, dest: Union[str, os.PathLike], is_dataset: bool = False) -> None:
            Protected method to handle model and dataset downloads.
        _get_url(alias: str, txt_path: str, use_bdl_api: bool, api_endpoint: str) -> str:
            Protected method to obtain the download URL from the BDL API or a text file.
    """

    def __init__(self):
        self.session = requests.Session()
        self.chunk_size = 1024
        self.BDL_MODEL_API = "https://balldatalab.com/api/models/"
        self.BDL_DATASET_API = "https://balldatalab.com/api/datasets/"
        self.yolo_model_aliases = {
            'phc_detector': 'models/YOLO/pitcher_hitter_catcher_detector/model_weights/pitcher_hitter_catcher_detector_v4.txt',
            'bat_tracking': 'models/YOLO/bat_tracking/model_weights/bat_tracking.txt',
            'ball_tracking': 'models/YOLO/ball_tracking/model_weights/ball_tracking.txt',
            'glove_tracking': 'models/YOLO/glove_tracking/model_weights/glove_tracking.txt',
            'ball_trackingv4': 'models/YOLO/ball_tracking/model_weights/ball_trackingv4.txt'
        }
        self.florence_model_aliases = {
            'ball_tracking': 'models/FLORENCE2/ball_tracking/model_weights/florence_ball_tracking.txt',
            'florence_ball_tracking': 'models/vlm/FLORENCE2/ball_tracking/model_weights/florence_ball_tracking.txt'
        }
        self.paligemma2_model_aliases = {
            'paligemma2_ball_tracking': 'models/vlm/paligemma2/ball_tracking/model_weights/paligemma2_ball_tracking.txt'
        }
        self.dataset_aliases = {
            'okd_nokd': 'datasets/yolo/OKD_NOKD.txt',
            'baseball_rubber_home_glove': 'datasets/yolo/baseball_rubber_home_glove.txt',
            'baseball_rubber_home': 'datasets/yolo/baseball_rubber_home.txt',
            'broadcast_10k_frames': 'datasets/raw_photos/broadcast_10k_frames.txt',
            'broadcast_15k_frames': 'datasets/raw_photos/broadcast_15k_frames.txt',
            'baseball_rubber_home_COCO': 'datasets/COCO/baseball_rubber_home_COCO.txt',
            'baseball_rubber_home_glove_COCO': 'datasets/COCO/baseball_rubber_home_glove_COCO.txt',
            'baseball': 'datasets/yolo/baseball.txt',
            'phc': 'datasets/yolo/phc.txt'
        }

    def _download_files(self, url: str, dest: Union[str, os.PathLike], is_folder: bool = False, is_labeled: bool = False) -> None:
        response = self.session.get(url, stream=True)
        if response.status_code == 200:
            total_size = int(response.headers.get('content-length', 0))
            progress_bar = tqdm(total=total_size, unit='iB', unit_scale=True, desc=f"Downloading {os.path.basename(dest)}")
            
            if is_folder: 
                content = io.BytesIO()
                for data in response.iter_content(chunk_size=self.chunk_size):
                    size = content.write(data)
                    progress_bar.update(size)
                
                progress_bar.close()

                with zipfile.ZipFile(content) as zip_ref:
                    for file in zip_ref.namelist():
                        if not file.startswith('__MACOSX') and not file.startswith('._'):
                            if is_labeled:
                                zip_ref.extract(file, dest)
                            else:
                                if '/' in file:
                                    filename = file.split('/')[-1]
                                    if filename:
                                        with zip_ref.open(file) as source, open(os.path.join(dest, filename), 'wb') as target:
                                            shutil.copyfileobj(source, target)
                                else:
                                    zip_ref.extract(file, dest)
                
                if not is_labeled:
                    for root, dirs, files in os.walk(dest, topdown=False):
                        for dir in dirs:
                            dir_path = os.path.join(root, dir)
                            if not os.listdir(dir_path):
                                os.rmdir(dir_path)
                
                print(f"Dataset downloaded and extracted to {dest}")
            else:
                with open(dest, 'wb') as file:
                    for data in response.iter_content(chunk_size=self.chunk_size):
                        size = file.write(data)
                        progress_bar.update(size)
                
                progress_bar.close()
                print(f"Model downloaded to {dest}")
        else:
            print(f"Download failed. STATUS: {response.status_code}")

    def _get_url(self, alias: str, txt_path: str, use_bdl_api: bool, api_endpoint: str) -> str:
        if use_bdl_api:
            return f"{api_endpoint}{alias}"
        else:
            with open(txt_path, 'r') as file:
                return file.read().strip()

    def load_model(self, model_alias: str, model_type: str = 'YOLO', use_bdl_api: Optional[bool] = True, model_txt_path: Optional[str] = None) -> str:
        '''
        Loads a given baseball computer vision model into the repository.

        Args:
            model_alias (str): Alias of the model to load.
            model_type (str): The type of the model to utilize. Defaults to YOLO.
            use_bdl_api (Optional[bool]): Whether to use the BallDataLab API.
            model_txt_path (Optional[str]): Path to .txt file containing download link to model weights. 
                                            Only used if use_bdl_api is specified as False.

        Returns:
            model_weights_path (str):  Path to where the model weights are saved within the repo.
        '''
        if model_type == 'YOLO':
            model_txt_path = self.yolo_model_aliases.get(model_alias) if use_bdl_api else model_txt_path
        elif model_type == 'FLORENCE2':
            model_txt_path = self.florence_model_aliases.get(model_alias) if use_bdl_api else model_txt_path
        elif model_type == 'PALIGEMMA2':
            model_txt_path = self.paligemma2_model_aliases.get(model_alias) if use_bdl_api else model_txt_path
        else:
            raise ValueError(f"Invalid model type: {model_type}")
        
        if not model_txt_path:
            raise ValueError(f"Invalid alias: {model_alias}")

        base_dir = os.path.dirname(model_txt_path)
        base_name = os.path.splitext(os.path.basename(model_txt_path))[0]
        os.makedirs(base_dir, exist_ok=True)

        if model_type == 'YOLO':
            model_weights_path = f"{base_dir}/{base_name}.pt"
        else:
            model_weights_path = f"{base_dir}/{base_name}"
            os.makedirs(model_weights_path, exist_ok=True)

        if os.path.exists(model_weights_path):
            print(f"Model found at {model_weights_path}")
            return model_weights_path

        url = self._get_url(model_alias, model_txt_path, use_bdl_api, self.BDL_MODEL_API)
        self._download_files(url, model_weights_path, is_folder=(model_type=='FLORENCE2'))
        
        return model_weights_path

    def load_dataset(self, dataset_alias: str, use_bdl_api: Optional[bool] = True, file_txt_path: Optional[str] = None) -> str:
        '''
        Loads a zipped dataset and extracts it to a folder.

        Args:
            dataset_alias (str): Alias of the dataset to load that corresponds to a dataset folder to download
            use_bdl_api (Optional[bool]): Whether to use the BallDataLab API. Defaults to True.
            file_txt_path (Optional[str]): Path to .txt file containing download link to zip file containing dataset. 
                                           Only used if use_bdl_api is specified as False.

        Returns:
            dir_name (str): Path to the folder containing the dataset.
        '''
        txt_path = self.dataset_aliases.get(dataset_alias) if use_bdl_api else file_txt_path
        if not txt_path:
            raise ValueError(f"Invalid alias or missing path: {dataset_alias}")

        base = os.path.splitext(os.path.basename(txt_path))[0]
        dir_name = "unlabeled_" + base if 'raw_photos' in base or 'frames' in base or 'frames' in dataset_alias else base

        if os.path.exists(dir_name):
            print(f"Dataset found at {dir_name}")
            return dir_name

        url = self._get_url(dataset_alias, txt_path, use_bdl_api, self.BDL_DATASET_API)
        os.makedirs(dir_name, exist_ok=True)
        self._download_files(url, dir_name, is_folder=True)

        return dir_name



import os
import subprocess
import sys
import cv2
import random

def extract_frames_from_video(video_path, game_id, output_frames_folder, frames_to_extract) -> list[str]:
    '''
    Extracts frames from a single video file and saves the images into the specified folder.
    '''
    cap = cv2.VideoCapture(video_path)

    if not cap.isOpened():
        print(f"Could not open video {os.path.basename(video_path)}...")
        return []

    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    frames_video = min(frames_to_extract, total_frames)
    frame_indices = sorted(random.sample(range(total_frames), frames_video))
    
    extracted_frames = []
    for i, frame_index in enumerate(frame_indices):
        cap.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
        ret, frame = cap.read()
        if ret:
            output_file = os.path.join(output_frames_folder, f"{game_id}_{os.path.splitext(os.path.basename(video_path))[0]}_{i:04d}.jpg")
            cv2.imwrite(output_file, frame)
            extracted_frames.append(output_file)
        else:
            print(f"Could not read frame {frame_index} from video {video_path}")
    
    cap.release()

    return extracted_frames

def check_import(install_path: str, package_name: str) -> bool:
    """
    Checks if a package is installed and attempts to install it if not found.

    Args:
        install_path (str): The path to the package to check.
        package_name (str): The name of the package to check.

    Returns:
        bool: True if the package is installed, False otherwise.
    """
    try:
        __import__(package_name)
        return True
    except ImportError:
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", install_path])
            return True
        except subprocess.CalledProcessError as e:
            print(f"Failed to install {package_name}: {str(e)}")
            raise
    





import pytest
import requests
from unittest.mock import patch

def test_bdl_api_call(mock_responses) -> None:
    """
    Tests the BDL API with both successful and error cases.
    
    Args:
        mock_responses (tuple): A tuple containing the success and error mock responses.
    """
    success_response, error_response = mock_responses
    
    with patch('requests.get') as mock_get:
        mock_get.side_effect = [success_response, error_response]

        # Test success
        response = requests.get('https://balldatalab.com/api/models/phc_detector')
        assert response.status_code == 200
        assert response.content == b"mock file content"
        assert "Content-Disposition" in response.headers
        
        # Test error
        with pytest.raises(requests.exceptions.HTTPError) as exc_info:
            response = requests.get('https://balldatalab.com/api/models/not_model')
            response.raise_for_status()
        
        assert exc_info.value.response.status_code == 404
        assert exc_info.value.response.json() == {"error": "File not found"}
        
        assert mock_get.call_count == 2
        mock_get.assert_any_call('https://balldatalab.com/api/models/phc_detector')
        mock_get.assert_any_call('https://balldatalab.com/api/models/not_model')




import os
import logging
from datetime import datetime
import sys

class ModelLogger:
    def __init__(self, model_name: str, model_run_path: str, model_id: str, batch_size: int, device: str) -> None:
        self.model_name = model_name
        self.model_run_path = model_run_path
        self.model_id = model_id
        self.batch_size = batch_size
        self.device = device

    def orig_logging(self) -> logging.Logger:
        """
        Set up logging for a given model.

        Returns:
            logging.Logger: The logger.
        """
        log_dir = os.path.join(self.model_run_path, "logs")
        os.makedirs(log_dir, exist_ok=True)

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        log_file = os.path.join(log_dir, f"{self.model_name}_{timestamp}.log")

        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
            handlers=[  
                logging.FileHandler(log_file),
                logging.StreamHandler(sys.stdout)
            ]
        )
        self.logger = logging.getLogger(f"{self.model_name}_({self.model_id})")
        self.logger.info(f"Initializing {self.model_name} model with Batch Size: {self.batch_size}")
        self.logger.info(f"Device: {self.device}")
        return self.logger



from .model_function_utils import ModelFunctionUtils
from .model_logger import ModelLogger
from .model_visualization_tools import ModelVisualizationTools

__all__ = ['ModelLogger', 'ModelFunctionUtils', 'ModelVisualizationTools']


from .processing.datasets_processor import DataProcessor
from .formats.datasets_coco_detection import CocoDetectionDataset
from .formats.datasets_jsonl_detection import JSONLDetection

__all__ = ['DataProcessor', 'CocoDetectionDataset', 'JSONLDetection']


import os
import shutil

def test_load_model(load_tools):
    """
    Tests loading a model using LoadTools with example call.
    Verifies that the model is downloaded correctly and the file is in the expected location.
    """
    os.makedirs('models/YOLO/pitcher_hitter_catcher_detector/model_weights', exist_ok=True)
    # Test loading PHC Detector model
    model_path = load_tools.load_model(
        model_alias="phc_detector",
        model_type="YOLO",
        use_bdl_api=True
    )
    
    assert os.path.exists(model_path), "Model file should exist after download"
    assert model_path.endswith('.pt'), "YOLO model should have .pt"
    
    if os.path.exists(model_path):
        os.remove(model_path)

def test_load_dataset(load_tools):
    """
    Tests loading a dataset using LoadTools with example call.
    Verifies that the dataset is downloaded and extracted correctly.
    """ 
    os.makedirs('datasets/baseball', exist_ok=True)
    # Test loading baseball dataset
    dataset_path = load_tools.load_dataset(
        dataset_alias="baseball",
        use_bdl_api=True
    )
    
    assert os.path.exists(dataset_path), "Dataset directory should exist"
    assert os.path.isdir(dataset_path), "Dataset path should be a directory"
    files = os.listdir(dataset_path)
    assert len(files) > 0, "Dataset should contain files"
    
    if os.path.exists(dataset_path):
        shutil.rmtree(dataset_path)


import os

import pytest

def test_run_statcast_pull_scraper(scraper):
    """
    Tests the main video scraping functionality.
    Uses yesterday's date to ensure game data exists and limits to 2 videos 
    total to keep the test quick while still verifying functionality.
    """
    #Run Scraper for Yankees on 3/28/2024 w/ minimal videos
    df = scraper.run_statcast_pull_scraper(
        start_date="2024-04-01",
        end_date="2024-04-01",
        download_folder="test_videos",
        max_workers=2,
        max_videos=2,
        max_videos_per_game=1,
        team="NYY"
    )
    
    assert not df.empty, "Should return data"
    assert os.path.exists("test_videos"), "Download folder should exist"
    videos = os.listdir("test_videos")
    assert len(videos) > 0 and videos[0].endswith(".mp4"), "Should have downloaded videos"
    scraper.cleanup_savant_videos("test_videos")

if __name__ == '__main__':
    pytest.main([__file__])


import logging
import cv2
import torch
from coco_eval import CocoEvaluator
from tqdm import tqdm
from transformers import DetrImageProcessor, DetrForObjectDetection
import os
import warnings
from datetime import datetime
from typing import Dict, List, Tuple
from torch.utils.data import DataLoader
import multiprocessing as mp
import supervision as sv
from baseballcv.model.utils import ModelLogger, ModelFunctionUtils, ModelVisualizationTools
from baseballcv.datasets import CocoDetectionDataset
import pytorch_lightning as pl

"""

DETR Class Implementation is based on the following tutorial:
- https://colab.research.google.com/github/roboflow-ai/notebooks/blob/main/notebooks/train-huggingface-detr-on-custom-dataset.ipynb

"""

class DETR(pl.LightningModule):
    """
    DETR (DEtection TRansformer) model class for object detection tasks using PyTorch Lightning and HuggingFace Transformers.
    Current class only supports COCO detection datasets.
    
    This class provides functionality for:
    - Model initialization and configuration
    - Fine-tuning on custom datasets
    - Inference on images and videos
    - Model evaluation
    
    Attributes:
        detr_device (torch.device): Device to run model on (cuda/mps/cpu)
        model_id (str): HuggingFace model identifier
        model_run_path (str): Directory path for model outputs
        model_name (str): Name of the model
        batch_size (int): Batch size for training/inference
        processor (DetrImageProcessor): Image processor for DETR
        model (DetrForObjectDetection): DETR model instance
        detr_logger (ModelLogger): Logger instance
        ModelFunctionUtils (ModelFunctionUtils): Utility functions for model operations
    """
    def __init__(self, 
                 num_labels: int,
                 device: str = None, 
                 model_id: str = "facebook/detr-resnet-50",
                 model_run_path: str = f'detr_run_{datetime.now().strftime("%Y%m%d")}',
                 batch_size: int = 8,
                 image_size: Tuple[int, int] = (800, 800)):
        """
        Initialize DETR model.

        Args:
            num_labels (int): Number of object classes to detect
            device (str, optional): Device to run model on. Defaults to None (auto-detect).
            model_id (str, optional): HuggingFace model ID. Defaults to "facebook/detr-resnet-50".
            model_run_path (str, optional): Output directory path. Defaults to timestamped directory.
            batch_size (int, optional): Batch size. Defaults to 8.
            image_size (Tuple[int, int], optional): Input image dimensions. Defaults to (800, 800).
        """
        super().__init__()
        self.save_hyperparameters(ignore=['device'])
        
        warnings.filterwarnings("ignore", category=FutureWarning, module="transformers")
        logging.getLogger("transformers").setLevel(logging.ERROR)
        
        # Device and Logger are both PL Lightning attributes - overriding with DETR specific attributes

        self.detr_device = "cuda" if torch.cuda.is_available() else "cpu" if device is None else device
        self.model_id = model_id
        self.model_run_path = model_run_path
        self.model_name = "DETR"
        self.batch_size = batch_size

        self.detr_logger = ModelLogger(self.model_name, self.model_run_path, 
                                self.model_id, self.batch_size, self.detr_device).orig_logging()

        self.processor = DetrImageProcessor.from_pretrained(
            self.model_id,
            size={"shortest_edge": image_size[0], "longest_edge": image_size[1]},
            do_resize=True,
            do_pad=True,
            do_normalize=True
        )
        
        self.model = DetrForObjectDetection.from_pretrained(
            self.model_id,
            num_labels=num_labels,
            ignore_mismatched_sizes=True
        ).to(self.detr_device)

        self.ModelFunctionUtils = ModelFunctionUtils(self.model_name, 
                            self.model_run_path, self.batch_size, 
                            self.detr_device, self.processor, self.model, 
                            None, self.detr_logger, torch.float16)
        
    def finetune(self, dataset_dir: str,
            classes: dict,
            save_dir: str = "finetuned_detr",
            batch_size: int = 4,
            epochs: int = 10,
            lr: float = 1e-4,
            lr_backbone: float = 1e-5,
            weight_decay: float = 0.01,
            gradient_accumulation_steps: int = 2,
            logging_steps: int = 100,
            save_limit: int = 2,
            patience: int = 3,
            patience_threshold: float = 0.0001,
            precision: str = '16-mixed',
            num_workers: int = None,
            freeze_layers: List[str] = None,
            freeze_head: bool = False,
            show_model_params: bool = True,
            push_to_hub_with_path: str = None,
            conf: float = 0.2) -> Dict:
        """
        Fine-tune the DETR model on a custom dataset.

        Args:
            dataset_dir (str): Path to dataset directory
            classes (dict): Dictionary mapping class IDs to class names
            save_dir (str, optional): Directory to save model. Defaults to "finetuned_detr".
            batch_size (int, optional): Training batch size. Defaults to 4.
            epochs (int, optional): Number of training epochs. Defaults to 10.
            lr (float, optional): Learning rate. Defaults to 1e-4.
            lr_backbone (float, optional): Backbone learning rate. Defaults to 1e-5.
            weight_decay (float, optional): Weight decay. Defaults to 0.01.
            gradient_accumulation_steps (int, optional): Gradient accumulation steps. Defaults to 2.
            logging_steps (int, optional): Steps between logging. Defaults to 100.
            save_limit (int, optional): Maximum checkpoints to save. Defaults to 2.
            patience (int, optional): Early stopping patience. Defaults to 3.
            patience_threshold (float, optional): Early stopping threshold. Defaults to 0.01.
            num_workers (int, optional): DataLoader workers. Defaults to None.
            freeze_layers (List[str], optional): Layers to freeze. Defaults to None.
            freeze_head (bool, optional): Whether to freeze head. Defaults to False.
            show_model_params (bool, optional): Show model parameters. Defaults to True.
            push_to_hub_with_path (str, optional): HuggingFace Hub path. Defaults to None.
            conf (float, optional): Confidence threshold. Defaults to 0.2.

        Returns:
            Dict: Training results...
                - best_model_path: Path to best checkpoint
                - best_model_score: Best validation score
                - early_stopped: Whether training stopped early
                - model_path: Path to saved model
        """
        try:
            self.model.config.id2label = classes
            self.lr = lr
            self.lr_backbone = lr_backbone
            self.weight_decay = weight_decay
            self.push_to_hub_with_path = push_to_hub_with_path

            model_path = os.path.join(self.model_run_path, save_dir, "model")
            os.makedirs(model_path, exist_ok=True)

            if freeze_head: 
                self.detr_logger.info("Freezing Head Layer...")
                self.ModelFunctionUtils.freeze_layers([
                    'model.backbone',
                    'model.encoder',
                    'model.decoder', 
                    'model.input_projection',
                    'model.query_position_embeddings',
                    'model.layernorm'
                ], show_params=show_model_params)

            elif freeze_layers is not None:
                self.ModelFunctionUtils.freeze_layers(freeze_layers, show_params=show_model_params)
            else:
                self.detr_logger.info("No Freezing Layers Specified - Training Entire Model")

            if not num_workers and self.detr_device != "mps":
                num_workers = min(12, mp.cpu_count() - 1)
                self.detr_logger.info(f"Using Default of {num_workers} workers")
            elif num_workers and num_workers > 0 and self.detr_device == "mps":
                num_workers = 0
                self.detr_logger.info("Using 0 workers for MPS")

            self.detr_logger.info("Loading Datasets...")
            train_dataset = CocoDetectionDataset(dataset_dir, "train", self.processor)
            val_dataset = CocoDetectionDataset(dataset_dir, "val", self.processor)

            train_loader = DataLoader(
                train_dataset,
                batch_size=batch_size,
                shuffle=True,
                num_workers=num_workers,
                collate_fn=self.collate_fn
            )
            
            val_loader = DataLoader(
                val_dataset,
                batch_size=batch_size,
                shuffle=False,
                num_workers=num_workers,
                collate_fn=self.collate_fn
            )

            callbacks = [
                pl.callbacks.ModelCheckpoint(
                    dirpath=os.path.join(model_path, 'checkpoints'),
                    filename='detr-{epoch:02d}-{validation_loss:.2f}',
                    monitor='validation_loss',
                    mode='min',
                    save_top_k=save_limit
                )
            ]

            if patience:
                callbacks.append(
                    pl.callbacks.EarlyStopping(
                        monitor='validation_loss',
                        patience=patience,
                        min_delta=patience_threshold,
                        mode='min'
                    )
                )

            trainer = pl.Trainer(
                max_epochs=epochs,
                accelerator='auto',
                devices=1,
                precision=precision,
                accumulate_grad_batches=gradient_accumulation_steps,
                callbacks=callbacks,
                logger=pl.loggers.TensorBoardLogger(
                    save_dir=model_path,
                    name=f'pl_tensorboard_logs_{self.model_name}'
                ),
                log_every_n_steps=logging_steps
            )

            self.detr_logger.info("Starting training...")
            trainer.fit(
                model=self,
                train_dataloaders=train_loader,
                val_dataloaders=val_loader
            )

            self.model.save_pretrained(model_path)
            self.processor.save_pretrained(model_path)
            self.detr_logger.info("Model and processor saved locally.")

            if self.push_to_hub_with_path is not None:
                self.detr_logger.info("Pushing to HuggingFace Hub...")

                try:
                    self.model.push_to_hub(self.push_to_hub_with_path)
                    self.processor.push_to_hub(self.push_to_hub_with_path)
                    self.detr_logger.info("Model and processor pushed to HuggingFace Hub successfully")
                except Exception as e:
                    self.detr_logger.error(f"Failed to push to HuggingFace Hub: {str(e)}")
            
            self.detr_logger.info("Training complete...")
            
            try:
                self.evaluate(dataset_dir=dataset_dir, conf=conf)
            except Exception as e:
                self.detr_logger.warning(f"Evaluation failed: {str(e)}")

            return {
                'best_model_path': trainer.checkpoint_callback.best_model_path,
                'best_model_score': trainer.checkpoint_callback.best_model_score,
                'early_stopped': trainer.early_stopping_callback.stopped_epoch > 0 if patience else False,
                'model_path': model_path
            }

        except Exception as e:
            self.detr_logger.error(f"Training failed: {str(e)}")
            raise e
        
    def inference(self, file_path: str,
                  classes: dict = None,
                  conf: float = 0.2, 
                  save: bool = False, 
                  save_viz_dir: str = 'visualizations',
                  show_video: bool = False) -> List[Dict]:
        """
        Run inference on image or video file.

        Args:
            file_path (str): Path to image or video file
            classes (dict, optional): Class ID to name mapping. Defaults to None.
            conf (float, optional): Confidence threshold. Defaults to 0.9.
            save (bool, optional): Whether to save visualizations. Defaults to False.
            save_viz_dir (str, optional): Directory to save visualizations. Defaults to 'visualizations'.
            show_video (bool, optional): Whether to display video during inference. Defaults to False.

        Returns:
            List[Dict]: List of detections, each containing:
                For images:
                    - boxes: Bounding box coordinates
                    - scores: Confidence scores
                    - labels: Predicted class labels
                For videos:
                    - video_id: Video file path
                    - frame: Frame number
                    - labels: Predicted class labels
                    - boxes: Bounding box coordinates
                    - scores: Confidence scores
        """

        if classes:
            self.model.config.id2label = classes

        self.model.eval()
        self.model.to(self.detr_device)
        if file_path.endswith(('.png', '.jpg', '.jpeg')):
            image = cv2.imread(file_path)

            with torch.no_grad():
                inputs = self.processor(images=image, return_tensors="pt").to(self.detr_device)
                outputs = self.model(**inputs)
                
                target_sizes = [(image.shape[:2])]
                results = self.processor.post_process_object_detection(
                    outputs,
                    target_sizes=target_sizes,
                    threshold=conf  
                )[0]
        
            detections = sv.Detections.from_transformers(transformers_results=results)
            labels = [
                f"{self.model.config.id2label[class_id]} {confidence:0.2f}" 
                for _, confidence, class_id, _ 
                in detections
            ]

            annotated_frame = ModelVisualizationTools(
                self.model_name, 
                self.model_run_path, 
                self.detr_logger
            ).visualize_detection_results(
                file_path=file_path, 
                detections=detections, 
                labels=labels, 
                save=save, 
                save_viz_dir=os.path.join(self.model_run_path, save_viz_dir)
            )

        elif file_path.endswith('.mp4'):
            cap = cv2.VideoCapture(file_path)
            width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            fps = cap.get(cv2.CAP_PROP_FPS)

            if save:
                save_path = os.path.join(self.model_run_path, save_viz_dir, f'{file_path.split("/")[-1]}.mp4')
                writer = cv2.VideoWriter(save_path, 
                               cv2.VideoWriter_fourcc(*'mp4v'),
                               fps, (width, height))
                
            all_detections = []
            frame_count = 0

            progress_bar = tqdm(total=total_frames, desc=f'Predicting Video: Frame {frame_count} / {total_frames}',
                                bar_format="{desc}\n{percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]\n{postfix}",
                                dynamic_ncols=True,
                                initial=0)

            with progress_bar as pbar:
                while cap.isOpened():
                    ret, frame = cap.read()
                    if not ret:
                        break

                    inputs = self.processor(images=image, return_tensors="pt")
                    inputs = {k: v.to(self.detr_device) for k, v in inputs.items()}
                    
                    with torch.no_grad():
                        outputs = self.model(**inputs)
                    
                    target_sizes = [(image.shape[:2])]
                    results = self.processor.post_process_object_detection(
                        outputs,
                        target_sizes=target_sizes,
                        threshold=conf  
                    )[0]
                    
                    detections = sv.Detections.from_transformers(transformers_results=results).with_nms(threshold=conf)
                    labels = [
                        f"{self.model.config.id2label[class_id]} {confidence:0.2f}" 
                        for _, confidence, class_id, _ 
                        in detections
                    ]
                    
                    frame_detections = []
                    for score, class_id, box in zip(detections.confidence, detections.class_id, detections.xyxy):
                        xmin, ymin, xmax, ymax = box.tolist()
                        detection = {
                            'video_id': file_path,
                            'frame': frame_count,
                            'labels': self.model.config.id2label[class_id],
                            'boxes': [xmin, ymin, xmax, ymax],  
                            'scores': score.item(),
                        }
                        frame_detections.append(detection)
                        annotated_frame = ModelVisualizationTools(
                            self.model_name, 
                            self.model_run_path, 
                            self.detr_logger
                        ).visualize_detection_results(
                            file_path=file_path, 
                            detections=detections, 
                            labels=labels, 
                            save=False, 
                            save_viz_dir=os.path.join(self.model_run_path, save_viz_dir)
                        )
                    
                    all_detections.extend(frame_detections)

                    if save:
                        writer.write(annotated_frame)

                    if show_video:
                        cv2.imshow('frame', annotated_frame)
                        if cv2.waitKey(1) & 0xFF == ord('q'):
                            break
                    
                    frame_count += 1
                    pbar.update(1)
            
   
        return detections if file_path.endswith(('.png', '.jpg', '.jpeg')) else all_detections


    def evaluate(self, dataset_dir: str,
                 conf: float = 0.2) -> Dict:
        """
        Evaluate model performance on test dataset.

        Args:
            dataset_dir (str): Path to dataset directory
            conf (float, optional): Confidence threshold. Defaults to 0.2.

        Returns:
            Dict: Evaluation metrics including mAP and other COCO metrics
        """
        try:
            self.detr_logger.info("Starting evaluation...")
            self.model.eval()
            self.model = self.model.to(self.detr_device)

            test_dataset = CocoDetectionDataset(dataset_dir, "test", self.processor)
            evaluator = CocoEvaluator(coco_gt=test_dataset.coco, iou_types=["bbox"])
            print(f"evaluator: {evaluator}")
            
            test_loader = DataLoader(
                test_dataset,
                batch_size=self.batch_size,
                shuffle=False,
                collate_fn=self.collate_fn
            )

            def convert_to_xywh(boxes):
                xmin, ymin, xmax, ymax = boxes.unbind(1)
                return torch.stack((xmin, ymin, xmax - xmin, ymax - ymin), dim=1)

            def prepare_for_coco_detection(predictions, conf):
                coco_results = []
                for original_id, prediction in predictions.items():
                    if len(prediction["boxes"]) == 0:
                        continue

                    boxes = convert_to_xywh(prediction["boxes"]).tolist()
                    scores = prediction["scores"].tolist()
                    labels = prediction["labels"].tolist()

                    for k, (box, score, label) in enumerate(zip(boxes, scores, labels)):
                        if score > conf:
                            coco_results.append({
                                "image_id": original_id,
                                "category_id": label,
                                "bbox": box,
                                "score": score,
                            })

                return coco_results

            for idx, batch in tqdm(enumerate(test_loader), desc="Evaluating", total=len(test_loader)):
                pixel_values = batch['pixel_values'].to(self.detr_device)
                pixel_mask = batch['pixel_mask'].to(self.detr_device)
                labels = [{k: v.to(self.detr_device) for k, v in t.items()} for t in batch['labels']]
                with torch.no_grad():
                    outputs = self.model(pixel_values=pixel_values, pixel_mask=pixel_mask)

                    orig_target_sizes = torch.stack([target["orig_size"] for target in labels], dim=0)
                    results = self.processor.post_process_object_detection(
                        outputs, 
                        target_sizes=orig_target_sizes, 
                        threshold=conf)
                    predictions = {target['image_id'].item(): output for target, output in zip(labels, results)}
                    coco_results = prepare_for_coco_detection(predictions, conf)
                    if coco_results:
                        evaluator.update(coco_results)

            evaluator.synchronize_between_processes()
            evaluator.accumulate()
            metrics = evaluator.summarize()

            return metrics

        except Exception as e:
            self.detr_logger.error(f"Evaluation failed: {str(e)}")
            raise e

    def forward(self, pixel_values, pixel_mask):
        """
        Forward pass of the model.

        Args:
            pixel_values: Input image pixel values
            pixel_mask: Input image pixel mask

        Returns:
            Model outputs
        """
        outputs = self.model(pixel_values=pixel_values, pixel_mask=pixel_mask)
        return outputs

    def common_step(self, batch, batch_idx):
        """
        Common step for training and validation.

        Args:
            batch: Input batch
            batch_idx: Batch index

        Returns:
            Tuple[torch.Tensor, Dict]: Loss and loss dictionary
        """
        pixel_values = batch["pixel_values"]
        pixel_mask = batch["pixel_mask"]
        labels = [{k: v.to(self.detr_device) for k, v in t.items()} for t in batch["labels"]]

        outputs = self.model(
            pixel_values=pixel_values,
            pixel_mask=pixel_mask,
            labels=labels
        )

        loss = outputs.loss
        loss_dict = outputs.loss_dict
        return loss, loss_dict

    def training_step(self, batch, batch_idx):
        """
        Training step.

        Args:
            batch: Input batch
            batch_idx: Batch index

        Returns:
            torch.Tensor: Training loss
        """
        loss, loss_dict = self.common_step(batch, batch_idx)
        
        self.log("training_loss", loss)
        for k, v in loss_dict.items():
            self.log("train_" + k, v.item())

        return loss

    def validation_step(self, batch, batch_idx):
        """
        Validation step.

        Args:
            batch: Input batch
            batch_idx: Batch index

        Returns:
            torch.Tensor: Validation loss
        """
        loss, loss_dict = self.common_step(batch, batch_idx)
        
        self.log("validation_loss", loss)
        for k, v in loss_dict.items():
            self.log("validation_" + k, v.item())

        return loss

    def configure_optimizers(self):
        """
        Configure optimizers for training.

        Returns:
            torch.optim.Optimizer: Configured optimizer
        """
        if not hasattr(self, 'lr') or not hasattr(self, 'lr_backbone') or not hasattr(self, 'weight_decay'):
            raise ValueError("Learning rates and weight decay must be set via finetune() before training")
        
        param_dicts = [
            {
                "params": [p for n, p in self.named_parameters() 
                          if "backbone" not in n and p.requires_grad]
            },
            {
                "params": [p for n, p in self.named_parameters() 
                          if "backbone" in n and p.requires_grad],
                "lr": self.lr_backbone,
            },
        ]
        
        optimizer = torch.optim.AdamW(
            param_dicts,
            lr=self.lr,
            weight_decay=self.weight_decay
        )
        
        return optimizer

    def collate_fn(self, batch):
        """
        Collate function for data loading.

        Args:
            batch: Input batch

        Returns:
            Dict: Collated batch with pixel values, mask and labels
        """
        pixel_values = [item[0] for item in batch]
        encoding = self.processor.pad(pixel_values, return_tensors="pt")
        labels = [item[1] for item in batch]
        return {
            'pixel_values': encoding['pixel_values'],
            'pixel_mask': encoding['pixel_mask'],
            'labels': labels
        }


import pandas as pd
from pandas.core.api import DataFrame
import requests
from concurrent.futures import ThreadPoolExecutor, as_completed
import os
from bs4 import BeautifulSoup
import time
import shutil
import polars as pl
import statcast_pitches

'''Class BaseballSavVideoScraper based on code from BSav_Scraper_Vid Repo, which can be found at https://github.com/dylandru/BSav_Scraper_Vid'''


class BaseballSavVideoScraper:
    def __init__(self):
        self.session = requests.Session()

    def run_statcast_pull_scraper(self,
                                  start_date: str | pd.Timestamp = '2024-05-01',
                                  end_date: str | pd.Timestamp = '2024-06-01',
                                  download_folder: str = 'savant_videos',
                                  max_workers: int = 5,
                                  team: str = None,
                                  pitch_call: str = None,
                                  max_videos: int = None,
                                  max_videos_per_game: int = None) -> pd.DataFrame:
        """
        Run scraper from Statcast Pull of Play IDs. Retrieves data and processes each row in parallel.

        Args:
            start_date (pd.Timestamp): Timestamp of start date for pull. Defaults to 2024-05-01.
            end_date (pd.Timestamp): Timestamp of end date for pull. Defaults to 2024-06-01.
            download_folder (str): Folder path where videos are downloaded. Defaults to 'savant_videos'.
            max_workers (int, optional): Max number of concurrent workers. Defaults to 5.
            team (str, optional): Team filter for which videos are scraped. Defaults to None.
            pitch_call (str, optional): Pitch call filter for which videos are scraped. Defaults to None.
            max_videos (int, optional): Max number of videos to pull. Defaults to None.
            max_videos_per_game (int, optional): Max number of videos to pull for single game. Defaults to None.

        Returns:
            pd.DataFrame: DataFrame containing Play IDs and relavent information on the videos that were scraped.

        Raises:
            Exception: Any error in downloading a video for a given play. 

        """
        try:
            print("Retrieving Play IDs to scrape...")
            df = self.playids_for_date_range(
                start_date=start_date, end_date=end_date, team=team, pitch_call=pitch_call)  # retrieves Play IDs to scrape

            if not df.empty and 'play_id' in df.columns:
                os.makedirs(download_folder, exist_ok=True)

                if max_videos_per_game is not None:
                    df = df.groupby('game_pk').head(
                        max_videos_per_game).reset_index(drop=True)

                if max_videos is not None:
                    df = df.head(max_videos)

                self.download_play_ids(download_folder, df, max_workers)
                return df

            else:
                print("Play ID column not in Statcast pull or DataFrame is empty")
                return pd.DataFrame()

        except KeyboardInterrupt:
            print("Ctrl+C detected. Shutting down.")
            return pd.DataFrame()

    def download_play_ids(self, download_folder: str, play_ids: DataFrame, max_workers: int = 5) -> DataFrame:
        """Download videos for given Play IDs."""
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            # sets futures to download videos for given Play IDs
            future_to_play_id = {executor.submit(
                self.get_video_for_play_id, row['play_id'], row['game_pk'], download_folder): row for _, row in play_ids.iterrows()}
            for future in as_completed(future_to_play_id):
                play_id = future_to_play_id[future]
                try:
                    future.result()
                except Exception as e:
                    print(
                        f"Error processing Play ID {play_id['play_id']}: {str(e)}")
            return play_ids

    def download_video(self, video_url, save_path, max_retries=5) -> None:
        """Downloads video from given URL and saves to specified path."""
        attempt = 0
        while attempt < max_retries:
            try:
                with self.session.get(video_url, stream=True) as r:
                    r.raise_for_status()
                    with open(save_path, 'wb') as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            f.write(chunk)
                return print(f"Video downloaded to {save_path}")
            except Exception as e:
                print(
                    f"Error downloading video {video_url}: {e}. Attempt {attempt + 1} of {max_retries}")
                attempt += 1
                time.sleep(2)

    def get_video_url(self, page_url, max_retries=5) -> str | None:
        """Retrieves Savant video URL from given page URL."""
        attempt = 0
        while attempt < max_retries:
            try:
                response = self.session.get(page_url)
                response.raise_for_status()
                soup = BeautifulSoup(response.content, 'html.parser')
                video_container = soup.find('div', class_='video-box')
                if video_container:
                    return video_container.find('video').find('source', type='video/mp4')['src']
                return None
            except Exception as e:
                print(
                    f"Error fetching video URL from {page_url}: {e}. Attempt {attempt + 1} of {max_retries}")
                attempt += 1
                time.sleep(2)

    def fetch_game_data(self, game_pk, max_retries=5):
        """Fetch game data for a single game_pk using the global session."""
        attempt = 0
        while attempt < max_retries:
            try:
                url = f'https://baseballsavant.mlb.com/gf?game_pk={game_pk}'
                response = self.session.get(url)
                response.raise_for_status()
                return response.json()
            except Exception as e:
                print(
                    f"Error fetching game data for game_pk {game_pk}: {e}. Attempt {attempt + 1} of {max_retries}")
                attempt += 1
                time.sleep(2)  # Wait for 2 seconds before retrying

    def process_game_data(self, game_data, pitch_call=None) -> DataFrame:
        """Process game data and filter by pitch_call if provided."""
        team_home_data = game_data.get('team_home', [])
        team_away_data = game_data.get('team_away', [])
        home_df = pd.json_normalize(team_home_data)
        away_df = pd.json_normalize(team_away_data)
        df = pd.concat([home_df, away_df], axis=0)
        if pitch_call:
            df = df.loc[df['pitch_call'] == pitch_call]
        return df

    def playids_for_date_range(self, start_date: str | pd.Timestamp, end_date: str | pd.Timestamp, team: str = None, pitch_call: str = None) -> DataFrame:
        """
        Retrieves PlayIDs for games played within date range. Can filter by team or pitch call.
        """
        start_date, end_date = pd.Timestamp(start_date) if isinstance(start_date, str) else start_date, pd.Timestamp(end_date) if isinstance(end_date, str) else end_date

        print(f"start_date: {start_date}")
        print(f"end_date: {end_date}")
        statcast_df = (statcast_pitches.load()
                       .filter(
            (pl.col("game_date").dt.date() >= start_date.date()) &
            (pl.col("game_date").dt.date() <= end_date.date()) &
            ((pl.col("home_team") == team) |
             (pl.col("away_team") == team)
             if team is not None else pl.lit(True)))
            .collect()
            .to_pandas())

        game_pks = statcast_df['game_pk'].unique()

        if len(game_pks) == 0:
            raise ValueError("No game_pks found for given date range and team")
        
        dfs = [self.process_game_data(self.fetch_game_data(
            game_pk), pitch_call=pitch_call) for game_pk in game_pks]
        
        if not dfs:
            raise ValueError("No data found for given date range and team")
        play_id_df = pd.concat(dfs, ignore_index=True)
        return play_id_df

    def get_video_for_play_id(self, play_id, game_pk, download_folder) -> None:
        """Process single play ID to download corresponding video."""
        page_url = f"https://baseballsavant.mlb.com/sporty-videos?playId={play_id}"
        try:
            video_url = self.get_video_url(page_url)
            if video_url:
                # Video currently named for Play ID
                save_path = os.path.join(
                    download_folder, f"{game_pk}_{play_id}.mp4")
                self.download_video(video_url, save_path)
            else:
                print(
                    f"No video found for playId {play_id}. Please check that the playId is correct or that the video exists at baseballsavant.mlb.com/sporty-videos?playId={play_id}.")
        except Exception as e:
            print(f"Unable to complete request. Error: {e}")

    def cleanup_savant_videos(self, folder_path: str) -> None:
        """Delete folder of downloaded BaseballSavant videos."""
        if os.path.exists(folder_path):
            try:
                shutil.rmtree(folder_path)
                print(f"Deleted {folder_path}")
            except Exception as e:
                print(f"Error deleting {folder_path}: {e}")

            return None



def test_imports():
    from baseballcv.functions import DataTools, LoadTools, BaseballSavVideoScraper
    from baseballcv.model import DETR, PaliGemma2, Florence2
    from baseballcv.datasets import CocoDetectionDataset, JSONLDetection, DataProcessor

    assert DataTools is not None
    assert LoadTools is not None
    assert BaseballSavVideoScraper is not None
    assert DETR is not None
    assert PaliGemma2 is not None
    assert Florence2 is not None
    assert CocoDetectionDataset is not None
    assert JSONLDetection is not None
    assert DataProcessor is not None


import os
import random
from collections import defaultdict
import concurrent.futures
from .savant_scraper import BaseballSavVideoScraper
from .utils import extract_frames_from_video
import shutil
from .load_tools import LoadTools
from ultralytics import YOLO
from tqdm import tqdm
from datetime import datetime

class DataTools:
    '''
    A class for generating and processing datasets for computer vision tasks in baseball.

    This class provides methods for generating photo datasets from baseball videos,
    and automating the annotation process using pre-trained models.

    Attributes:
        scraper (BaseballSavVideoScraper): An instance of the BaseballSavVideoScraper class.
        process_pool (concurrent.futures.ProcessPoolExecutor): A process pool for parallel execution.
        LoadTools (LoadTools): An instance of the LoadTools class for loading models.
        output_folder (str): The output folder for generated datasets.
    '''

    def __init__(self, max_workers: int = 10):
        self.scraper = BaseballSavVideoScraper()
        self.max_workers = max_workers
        self.LoadTools = LoadTools()
        self.output_folder = ''

    def generate_photo_dataset(self,
                           output_frames_folder: str = "cv_dataset", 
                           video_download_folder: str = "raw_videos",
                           max_plays: int = 10, 
                           max_num_frames: int = 6000,
                           max_videos_per_game: int = 10,
                           start_date: str = "2024-05-22",
                           end_date: str = "2024-07-25",
                           delete_savant_videos: bool = True) -> None:
        """
        Extracts random frames from scraped Baseball Savant broadcast videos to create a photo dataset for a 
        Computer Vision model.
        
        Args:
            output_frames_folder (str): Name of folder where photos will be saved. Default is "cv_dataset".
            video_download_folder (str): Name of folder containing videos. Default is "raw_videos".
            max_plays (int): Maximum number of plays for scraper to download videos. Default is 10.
            max_num_frames (int): Maximum number of frames to extract across all videos. Default is 6000.
            max_videos_per_game (int): Max number of videos to pull for single game to increase variety. Defaults to 10.
            start_date (str): Start date for video scraping in "YYYY-MM-DD" format. Default is "2024-05-22".
            end_date (str): End date for video scraping in "YYYY-MM-DD" format. Default is "2024-05-25".
            max_workers (int): Number of worker processes to use for frame extraction. Default is 10.
            delete_savant_videos (bool): Whether or not to delete scraped savant videos after frames are extracted. Default is True.

        Returns:
            None: Creates a folder of photos from the video frames to use.
        """

        self.output_folder = output_frames_folder
        start_date = datetime.strptime(start_date, "%Y-%m-%d")
        end_date = datetime.strptime(end_date, "%Y-%m-%d")
        self.scraper.run_statcast_pull_scraper(start_date=start_date, end_date=end_date, 
                                download_folder=video_download_folder, max_videos=max_plays, max_videos_per_game=max_videos_per_game, max_workers=1)
                
        os.makedirs(self.output_folder, exist_ok=True)
        video_files = [f for f in os.listdir(video_download_folder) if f.endswith(('.mp4', '.mov'))]
        
        if not video_files:
            print("No video files found in the specified folder.") 
            return
        
        games = defaultdict(list) #group videos by given game for increased variety
        for video_file in video_files:
            game_id = video_file[:6] 
            games[game_id].append(video_file)
        
        frames_per_game = max_num_frames // len(games)
        remaining_frames = max_num_frames % len(games) #distribute frames evenly across games
        
        extraction_tasks = []
        for game_id, game_videos in games.items():
            frames_for_game = frames_per_game + (1 if remaining_frames > 0 else 0)
            remaining_frames = max(0, remaining_frames - 1)
            
            frames_per_video = frames_for_game // len(game_videos)
            extra_frames = frames_for_game % len(game_videos)
            
            for i, video_file in enumerate(game_videos):
                frames_to_extract = frames_per_video + (1 if i < extra_frames else 0)
                video_path = f"{video_download_folder}/{video_file}"
                extraction_tasks.append((video_path, game_id, self.output_folder, frames_to_extract))
        
        extracted_frames = []

        with concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers) as executor:
            try:
                future_videos = {executor.submit(extract_frames_from_video, *task): task 
                               for task in extraction_tasks}
                for future in concurrent.futures.as_completed(future_videos):
                    video_path, game_id, _, _ = future_videos[future]
                    try:
                        result = future.result()
                        extracted_frames.extend(result)
                    except Exception as e:
                        print(f"Error with {video_path}: {str(e)}")
            except KeyboardInterrupt:
                print("\nShutting down gracefully...")
                executor.shutdown(wait=False)
                raise
            finally:
                for task in extraction_tasks:
                    video_path = task[0]
                    if os.path.exists(video_path):
                        try:
                            os.close(os.open(video_path, os.O_RDONLY))
                        except:
                            pass

        random.shuffle(extracted_frames)
        
        for i, frame_path in enumerate(extracted_frames):
            frame_name = f"{i+1:06d}{os.path.splitext(frame_path)[1]}"
            new_path = os.path.join(self.output_folder, frame_name)
            shutil.move(frame_path, new_path) 

        existing_files = set(os.listdir(self.output_folder))
        extracted_file_names = set(f"{i+1:06d}{os.path.splitext(frame)[1]}" for i, frame in enumerate(extracted_frames))
        files_to_remove = existing_files - extracted_file_names
        for file in files_to_remove:
            os.remove(os.path.join(self.output_folder, file))
        
        print(f"Extracted {len(extracted_frames)} frames from {len(video_files)} videos over {len(games)} games.")
        
        if delete_savant_videos:
            self.scraper.cleanup_savant_videos(video_download_folder)
    
    def automated_annotation(self, 
                             model_alias: str,
                             model_type: str = 'detection',
                             image_dir: str = "cv_dataset",
                             output_dir: str = "labeled_dataset", 
                             conf: float = .80, 
                             device: str = 'cpu') -> None:
        """
        Automatically annotates images using pre-trained YOLO model from BaseballCV repo. The annotated output
        consists of image files in the output directory, and label files in the subfolder "annotations" to 
        work with annotation tools.

        Note: The current implementation only supports YOLO detection models. 

        Args:
            model_alias (str): Alias of model to utilize for annotation.
            model_type (str): Type of CV model to utilize for annotation. Default is 'detection'.
            image_dir (str): Directory with images to annotate. Default is "cv_dataset".
            output_dir (str): Directory to save annotated images / labels. Default is "labeled_dataset".
            conf (float): Minimum confidence threshold for detections. Default is 0.80.
            device (str): Device to run model on ('cpu', 'mps', 'cuda'). Default is 'cpu'.

        Returns:
            None: Saves annotated images and labels to the output directory.
        """
        os.makedirs(output_dir, exist_ok=True)
        annotations_dir = os.path.join(output_dir, "annotations")
        os.makedirs(annotations_dir, exist_ok=True)

        model = YOLO(self.LoadTools.load_model(model_alias))
        print(f"Model loaded: {model}")

        annotation_tasks = [image_file for image_file in os.listdir(image_dir)]

        for image_file in tqdm(annotation_tasks, desc="Annotating images"):
            image_path = os.path.join(image_dir, image_file)
            annotations = []

            results = model.predict(source=image_path, save=False, conf=conf, device=device, verbose=False)
            
            if model_type == 'detection':
                for result in results:
                    for box in result.boxes:
                        cls = int(box.cls)
                        xywhn = box.xywhn[0].tolist()
                        if len(xywhn) == 4:
                            x_center, y_center, width, height = xywhn
                            annotations.append(f"{cls} {x_center} {y_center} {width} {height}")
                        else:
                            print(f"Invalid bounding box for {image_file}: {xywhn}")

            #TODO: Add annotation format for YOLO Keypoint, Segmentation, and Classification models

            if annotations:
                shutil.copy(image_path, output_dir)
                output_file = os.path.join(annotations_dir, os.path.splitext(image_file)[0] + '.txt')
                os.makedirs(os.path.dirname(output_file), exist_ok=True)
                with open(output_file, 'w') as f:
                    f.write('\n'.join(annotations))

        print("Annotation process complete.")

        return None




import os
import torch
from transformers import DetrImageProcessor
import torchvision
from pycocotools.coco import COCO

class CocoDetectionDataset(torchvision.datasets.CocoDetection):
    """
    Dataset class for COCO-format detection datasets that supports both:
    1. Hierarchical structure with split/images and split/labels
    2. Flat structure with all images and instances_{split}.json files in root dir
    """
    def __init__(self, dataset_dir: str, split: str, processor: DetrImageProcessor):
        """
        Initialize COCO detection dataset.
        
        Args:
            dataset_dir (str): Root directory containing the dataset
            split (str): Dataset split ('train', 'test', or 'val')
            processor (DetrImageProcessor): DETR image processor for preprocessing
        """
        
        # Check for hierarchical structure first
        hierarchical_img_dir = os.path.join(dataset_dir, split)
        hierarchical_ann_file = os.path.join(dataset_dir, split, "_coco.json")
        
        # Check for flat structure
        flat_img_dir = dataset_dir
        flat_ann_file = os.path.join(dataset_dir, f"instances_{split}.json")
        
        # Determine which structure to use
        if os.path.exists(hierarchical_img_dir) and os.path.exists(hierarchical_ann_file):
            self.img_dir = hierarchical_img_dir
            self.ann_file = hierarchical_ann_file
        elif os.path.exists(flat_img_dir) and os.path.exists(flat_ann_file):
            self.img_dir = flat_img_dir
            self.ann_file = flat_ann_file
        else:
            raise ValueError(
                f"Could not find valid COCO dataset structure in {dataset_dir}. "
                f"Expected either:\n"
                f"1. Hierarchical: {hierarchical_img_dir} and {hierarchical_ann_file}\n"
                f"2. Flat: {flat_img_dir} and {flat_ann_file}"
            )
            
        super(CocoDetectionDataset, self).__init__(self.img_dir, self.ann_file)
        self.processor = processor

    def __getitem__(self, idx):
        images, annotations = super(CocoDetectionDataset, self).__getitem__(idx)        
        image_id = self.ids[idx]
        annotations = {'image_id': image_id, 'annotations': annotations}
        encoding = self.processor(images=images, annotations=annotations, return_tensors="pt")
        pixel_values = encoding["pixel_values"].squeeze()
        target = encoding["labels"][0]

        return pixel_values, target



import random
import logging
from typing import List, Dict, Any, Tuple
from PIL import Image, ImageEnhance
import os
import json
from torch.utils.data import Dataset
import numpy as np

class JSONLDetection(Dataset):
    '''
    A class to load YOLO datasets into a JSONL format that can be used for training and validation in PyTorch.
    '''
    def __init__(self, entries: List[Dict[str, Any]], image_directory_path: str, logger: logging.Logger, augment: bool = True) -> None:
        """
        Initialize the JSONLDetection dataset.

        Args:
            entries (List[Dict[str, Any]]): List of entries (annotations) for the dataset.
            image_directory_path (str): Path to the directory containing images.
            logger (logging.Logger): Logger instance for logging.
            augment (bool): Whether to apply data augmentation.
        """
        self.entries = entries
        self.image_directory_path = image_directory_path
        self.augment = augment
        self.transforms = [self.random_color_jitter, self.random_blur, self.random_noise] if augment else []
        self.logger = logger

    def __len__(self) -> int:
        """
        Return the number of samples in the dataset.

        Returns:
            int: The number of samples in the dataset.
        """
        return len(self.entries)

    def __getitem__(self, idx: int) -> Tuple[Image.Image, Dict[str, Any]]:
        """
        Get a sample from the dataset.

        Args:
            idx (int): Index of the sample to retrieve.

        Returns:
            image (PIL.Image.Image): The image.
            entry (Dict[str, Any]): The entry.
        """
        entry = self.entries[idx]
        image_path = os.path.join(self.image_directory_path, entry['image'])
        image = Image.open(image_path)

        if self.augment and random.random() > 0.5:
            for transform in self.transforms:
                image = transform(image)

        return image, entry
    
    @staticmethod
    def random_color_jitter(image: Image.Image) -> Image.Image:
        """
        Apply random color jittering to image.

        Args:
            image (PIL.Image.Image): The image to apply the transformation to.

        Returns:
            image (PIL.Image.Image): The transformed image.
        """
        factors = {
            'brightness': random.uniform(0.8, 1.2),
            'contrast': random.uniform(0.8, 1.2),
            'color': random.uniform(0.8, 1.2)
        }

        for enhance_type, factor in factors.items():
            if random.random() > 0.5:
                if enhance_type == 'brightness':
                    image = ImageEnhance.Brightness(image).enhance(factor)
                elif enhance_type == 'contrast':
                    image = ImageEnhance.Contrast(image).enhance(factor)
                elif enhance_type == 'color':
                    image = ImageEnhance.Color(image).enhance(factor)
        return image

    @staticmethod
    def random_blur(image: Image.Image) -> Image.Image:
        """
        Apply random Gaussian blur to the image.

        Args:
            image (PIL.Image.Image): The image to apply the transformation to.

        Returns:
            image (PIL.Image.Image): The transformed image.
        """
        if random.random() > 0.8:
            from PIL import ImageFilter
            return image.filter(ImageFilter.GaussianBlur(radius=random.uniform(0.5, 1.0)))
        return image

    @staticmethod
    def random_noise(image: Image.Image) -> Image.Image:
        """
        Apply random noise to the image.

        Args:
            image (PIL.Image.Image): The image to apply the transformation to.

        Returns:
            image (PIL.Image.Image): The transformed image.
        """
        if random.random() > 0.8:
            img_array = np.array(image)
            noise = np.random.normal(0, 2, img_array.shape)
            noisy_img = np.clip(img_array + noise, 0, 255).astype(np.uint8)
            return Image.fromarray(noisy_img)
        return image
    
    @staticmethod
    def load_jsonl_entries(jsonl_file_path: str, logger: logging.Logger) -> List[Dict[str, Any]]:
        """
        Load entries from a JSONL file.

        Args:
            jsonl_file_path (str): Path to the JSONL file.
            logger (logging.Logger): Logger instance for logging.

        Returns:
            List[Dict[str, Any]]: List of entries loaded from the JSONL file.
        
        Raises:
            ValueError: If no valid entries are found in the JSONL file.
            Exception: If there is an error loading entries from the JSONL file.
        """
        entries = []
        try:
            with open(jsonl_file_path, 'r') as file:
                for line in file:
                    try:
                        data = json.loads(line.strip())
                        if isinstance(data, dict):
                            entries.append(data)
                        else:
                            logger.warning(f"Skipping invalid entry: not dictionary - {data}")
                    except json.JSONDecodeError as e:
                        logger.warning(f"Failed to parse JSON line: {e}")
                        continue

            if not entries:
                logger.error(f"No valid entries found in {jsonl_file_path}")
                raise ValueError(f"No valid entries found in {jsonl_file_path}")

            logger.info(f"Loaded {len(entries)} valid entries from {jsonl_file_path}")
            return entries

        except Exception as e:
            logger.error(f"Error loading entries from {jsonl_file_path}: {str(e)}")
            raise



import shutil
import os
import pytest

def test_generate_photo_dataset(data_tools):
    """
    Tests the generate_photo_dataset method using example call.
    Keeps part of output for the following test_automated_annotation method.
    """
    
    # Use small dates and limited plays for testing
    try:
        data_tools.generate_photo_dataset(
            output_frames_folder="test_dataset",
            video_download_folder="test_videos",
            max_plays=2,
            max_num_frames=10,
            max_videos_per_game=1,
            start_date="2024-04-01",
            end_date="2024-04-01",
            delete_savant_videos=True
        )
        
        assert os.path.exists("test_dataset")
        frames = os.listdir("test_dataset")
        assert len(frames) > 0, "Should have generated at least some frames"
        
        for frame in frames:
            assert frame.endswith(('.jpg', '.png')), f"Invalid frame format: {frame}"
    
    except Exception as e:
        print(f"An error occurred: {e}")
        raise
            

def test_automated_annotation(load_tools, data_tools):
    """
    Tests the automated_annotation method using example call.
    """
    test_generate_photo_dataset(data_tools)  # Reuse dataset generation
    
    try:
        os.makedirs('models/YOLO/glove_tracking/model_weights', exist_ok=True)
        data_tools.automated_annotation(
            model_alias="glove_tracking",  
            model_type="detection",
            image_dir="test_dataset",
            output_dir="test_annotated",
            conf=0.5
        )
        
        assert os.path.exists("test_annotated")
        assert os.path.exists(os.path.join("test_annotated", "annotations"))
        
        images = os.listdir("test_annotated")
        annotations = os.listdir(os.path.join("test_annotated", "annotations"))
        assert len(images) > 0, "Should have copied some images"
        assert len(annotations) > 0, "Should have generated some annotations"
        
        for ann_file in annotations:
            assert ann_file.endswith('.txt'), f"Invalid annotation format: {ann_file}"
        
        os.remove(load_tools.yolo_model_aliases['glove_tracking'].replace('.txt', '.pt'))
        shutil.rmtree("test_annotated")
        shutil.rmtree("test_dataset")

    except Exception as e:
        print(f"An error occurred: {e}")
        raise

if __name__ == '__main__':
    pytest.main([__file__])


import os
import cv2
import matplotlib.pyplot as plt
from typing import List
import logging
import supervision as sv


class ModelVisualizationTools:
    '''
    A class to visualize the results of a model.
    '''
    def __init__(self, model_name: str, model_run_path: str, logger: logging.Logger) -> None:
        """
        Initialize the ModelVisualizationTools class.

        Args:
            model_name (str): The name of the model.
            model_run_path (str): The path to the model run directory.
            logger (logging.Logger): The logger instance for logging.
        """
        self.model_name = model_name
        self.model_run_path = model_run_path
        self.logger = logger

    def visualize_detection_results(self, file_path: str, detections: sv.Detections, labels: List[str], save: bool = True, save_viz_dir: str = 'visualizations') -> logging.Logger:
        """
        Visualize the results.

        Args:
            file_path (str): The path to the image file.
            detections (sv.Detections): Detections object containing boxes, class IDs and confidence scores
            labels (List[str]): List of formatted label strings
            save_viz_dir (str): Directory to save the visualizations.

        Returns:
            logger_message (logging.Logger): The logger message for logging the completed visualization saving.
        """

        image = cv2.imread(file_path)
        box_annotator = sv.BoxAnnotator()
        annotated_frame = box_annotator.annotate(scene=image, detections=detections, labels=labels)

        fig, ax = plt.subplots(figsize=(10, 8))
        ax.imshow(cv2.cvtColor(annotated_frame, cv2.COLOR_BGR2RGB))
        ax.axis('off')

        if save:
            os.makedirs(save_viz_dir, exist_ok=True)
            
            if isinstance(file_path, str):
                base_name = os.path.basename(file_path)
                save_name = os.path.splitext(base_name)[0]
            else:
                save_name = 'frame'
                
            save_path = os.path.join(save_viz_dir, f'detection_{save_name}.jpg')
            
            cv2.imwrite(save_path, annotated_frame)
            self.logger.info(f"Visualizations saved to {save_path}")
        
        plt.show()
        plt.close()

        return (save_path, annotated_frame) if save else annotated_frame


        

    